-- Common stuff

if not srb2gb then rawset(_G, "srb2gb", {}) end

--Time management(only for 2.2.10)
local haveTime = true
local timeHist = 0
local timeHistE = 1
if not getTimeMicros then
	haveTime = false
	CONS_Printf(consoleplayer, "Require SRB2 2.2.10 or newer. It will work but some features may not work correctly")
	getTimeMicros = function() return -1 end
end
local dtime = 0
local speedFac = FRACUNIT
local speedMul = FRACUNIT

local emu = srb2gb

emu.debug = CV_RegisterVar({
	name = "emudebug",
	defaultvalue = "no",
	PossibleValue = CV_YesNo
})

local autoslow = CV_RegisterVar({ --Skip tic if nothing was drawn last tic(to avoid decreasing fps on weak computer)
	name = "emulimiter",
	defaultvalue = "14000",
	PossibleValue = CV_Unsigned
})

emu.drawn = false --Did a frame was drawn last tic?

local wram = {}
emu.wram = wram
local hram = {}
emu.hram = hram

emu.rate = 0x100000 --1MiHz
local state = 0
emu.reg = {}
local reg = emu.reg
local memlock = false
local stop = 0
--MMIO functions
if not emu.mmioW then
	emu.mmioW = {}
	emu.mmioR = {}
end
local ine = 0x00 --Interrupt enable. What interrupt work
local inf = 0 --Interrupt flags. If a bit is set in both this and ine is set call an interrupt vector
local pc -- Don't expose it in an attempt to increase speed(even if the most called instruction is JR)
local ime = false

local divTic = 0
local MMIODIV = 4

local timer = 0
local timerMod = 0
local timerCtrl = 0xF8
local TIMERENABLE = 0x4
local TIMERCLOCK = 0x3
local timerClockFreq = {255, 3, 15, 63} -- Offset to accomodate offset timer
local currentFreq = 255 -- Don't check the array each time 
local MMIOTIMA = 5
local MMIOTMA = 6
local MMIOTAC = 7

local MMIOIF = 0xF

local WRAMMASK = 0x1FFF

local REGSP = emu.REGSP
local REGAF = emu.REGAF
local REGAFBIS = REGAF+1

local cpuCycle = 0
local cyclePerTic = emu.rate / TICRATE --Emulation speed of 100%

local intVector = 0x40 --First interrupt vector(offset by 8 for each interrupt tested)

local ROMCHECKSUM = 0x014D --Initial flag values depends on this... At least no need to put bootrom

local STATERUN = emu.STATERUN
local STOPSTOP = emu.STOPSTOP
local STOPHALT = emu.STOPHALT

local memR = {} -- Lua would work better with BSP switch. Still better than if chain on namespace function
emu.memR = memR -- Exposed for init(and dynamic ROM function)

emu.speed = CV_RegisterVar({
	name = "emuspeed",
	defaultvalue = "0",
	flags = CV_FLOAT,
	PossibleValue = CV_Unsigned
})

local toInt = function(val)
	if val == nil then return 0 end
	return val&0xFF
end

local readWRAM = function(addr)
	return wram[(addr&WRAMMASK)]
end
emu.readWRAM = readWRAM --Expose it for dma and stuff

memR[0xC] = readWRAM
memR[0xD] = readWRAM
memR[0xE] = readWRAM

local writeWRAM = function(addr, val)
	wram[(addr&WRAMMASK)] = val
end

--Why don't I use string.format?
emu.hexString = function(val, size)
	if size == nil then size = 8 end
	local n
	local s = ""
	for i = 1, size do --For each nibble nibbles
		n = val & 15
		val = $ >> 4
		if n < 10 then -- 0-9
			s = tostring(n)..$
		else -- A-F
			s = string.char(97 + (n - 10))..$
		end
	end
	return s
end

emu.init = function()
	--Init register to DMG values
	reg[emu.REGAF + 1] = 0x0180
	reg[emu.REGBC + 1] = 0x0013
	reg[emu.REGDE + 1] = 0x00D8
	reg[emu.REGHL + 1] = 0x014D
	pc = 0x100
	emu.sp = 0xFFFE
	ine = 0
	ime = false
	inf = 0xE5
	divTic = 0x2AF3 -- Yes this feel random but what I found on doc
	
	stop = emu.STOPNO
	
	memlock = false
	
	--Set flags to initial values. Some depends on checksum
	local c = emu.romRead(ROMCHECKSUM)
	if c then emu.reg[emu.REGAF + 1] = $ | emu.FLAGCMASK | emu.FLAGHMASK end
		
	
	--Fill WRAM and HRAM with 0
	for i = 0, 0x1FFF do
		wram[i] = 0
	end
	for i = 0, 126 do
		hram[i] = 0
	end
	
	emu.initLCD()
end

emu.start = function(rom)
	--Start emulation
	if emu.loadRom(rom) then
		emu.init()
		state = emu.STATERUN
		if emu.debug.value then
			state = emu.STATEERROR
		end
		return true
	else
		state = emu.STATEERROR
		print("Failed to open ROM \""..rom.."\"")
	end
	return false
end

emu.setState = function(st)
	state = st
end

emu.getState = function()
	return state
end

--Timing
local nextCycle = function()
	cpuCycle = $ - 1
	emu.lCDstep()
	divTic = ($ + 1)&0x3FFF -- Timer use the full clock it seems. But nothing use the full byte so should be safe to assume it increase once
	--The timer
	if timerCtrl & TIMERENABLE and not (divTic & currentFreq) then
		timer = ($ + 1)
		if timer > 0xFF then
			timer = timerMod
			emu.callInt(emu.INTTIMER)
		end
	end
end

emu.nextCycle = nextCycle

emu.addCycle = function(nb)
	for i = 1, nb do
		emu.nextCycle()
	end
end

--Read a byte
local read8 = function(addr) -- If/else only on >= 0xF000 should be a bit faster(I hope)
	nextCycle()
	if addr < 0xFF00 and memlock then return 0 end
	if addr < 0xF000 then
		return memR[addr >> 12](addr)
	end
	--Let emulate memory map
	if addr < 0xFE00 then -- C000 - FDFF : WRAM + Echo
		return readWRAM(addr)
	elseif addr < 0xFF00 then -- FE00 - FE9F : OAM(and put invalid range in OAM)
		return emu.readOAM(addr)
	elseif addr < 0xFF80 then -- FF00 - FF7F : MMIO
		if emu.mmioR[addr&0x7F] then
			return emu.mmioR[addr&0x7F]()
		elseif emu.debug.value then
			print("Read MMIO: "..emu.hexString(addr,4).." "..emu.hexString(emu.pc,4))
		end
		return 0xFF
	elseif addr < 0xFFFF then
		return hram[(addr&0x7F)]
	elseif addr == 0xFFFF then
		return ine
	end
	error("Read at out of range address", 2)
	return 0xFF
end

local read8Raw = function(addr) -- For dump
	addr = $ & 0xFFFF
	if addr < 0xF000 then
		return memR[addr >> 12](addr)
	end
	--Let emulate memory map
	if addr < 0xFE00 then -- C000 - FDFF : WRAM + Echo
		return readWRAM(addr)
	elseif addr < 0xFF00 then -- FE00 - FE9F : OAM(and put invalid range in OAM)
		return emu.readOAM(addr)
	elseif addr < 0xFF80 then -- FF00 - FF7F : MMIO
		if emu.mmioR[addr&0x7F] then
			return emu.mmioR[addr&0x7F]()
		elseif emu.debug.value then
			print("Read MMIO: "..emu.hexString(addr,4).." "..emu.hexString(emu.pc,4))
		end
		return 0xFF
	elseif addr < 0xFFFF then
		return hram[(addr&0x7F)]
	elseif addr == 0xFFFF then
		return ine
	end
	return 0xFF
end

emu.read8 = read8

local read16Raw = function(addr)
	return read8Raw(addr) + (read8Raw(addr + 1) << 8)
end

local read16 = function(addr)
	return read8(addr) + (read8(addr + 1) << 8)
end
emu.read16 = read16

local readNext8 = function()
	local v = read8(pc) 
	pc = $ + 1
	return v
end
emu.readNext8 = readNext8

emu.readNext16 = function()
	local v = read16(pc)
	pc = $ + 2
	return v
end

--Write
local write8 = function(addr, val)
	nextCycle()
	if addr < 0xFF00 and memlock then return end
	val = $ & 0xFF
	--Let emulate memory map
	if addr < 0x8000 then -- 0-7FFF : ROM
		emu.romWrite(addr, val)
	elseif addr < 0xA000 then -- 8000-9FFF : VRAM
		emu.writeVRAM(addr, val)
	elseif addr < 0xC000 then -- A000 - BFFF : SRAM
		emu.sramWrite(addr, val)
	elseif addr < 0xFE00 then -- C000 - DFFF : WRAM + Echo
		writeWRAM(addr, val)
	elseif addr < 0xFEA0 then -- FE00 - FE9F : OAM
		emu.writeOAM(addr, val)
	elseif addr < 0xFF00 then -- FEA0 - FEFF : Unusable
		--Do nothing
	elseif addr < 0xFF80 then -- FF00 - FF7F : MMIO
		if emu.mmioW[addr&0x7F] then
			emu.mmioW[addr&0x7F](val)
		end
	elseif addr == 0xFFFF then
		ine = val
	else
		hram[(addr&0x7F)] = val
	end
end

emu.write8 = write8

--Reg
local getRegPair = function(r)
	r = $ & 3
	return reg[r + 1]
end
emu.getRegPair = getRegPair

emu.getReg = function(r)
	r = r & 7
	local val = reg[(r >> 1) + 1]
	if (not (r&1)) or r == 7 then val = $ >> 8 end -- Why A don't follow the same rules
	return val&0xFF
end

local setRegPair = function(r, v)
	v = v & 0xFFFF
	reg[(r&3) + 1] = v
end
emu.setRegPair = setRegPair

emu.setReg = function(r, v)
	v = v & 0xFF
	r = r & 7
	if r&1 and r ~= 7 then
		reg[(r >> 1) + 1] = ($&0xFF00) | v
	else
		reg[(r >> 1) + 1] = ($&0xFF) | (v << 8)
	end
end

emu.setFlag = function(flag, val)
	if val then
		reg[REGAFBIS] = $ | flag
	else
		reg[REGAFBIS] = $ & (~flag)
	end
end

local push = function(val)
	write8(emu.sp - 1, (val >> 8))
	write8(emu.sp - 2, val)
	emu.sp = $ - 2
end
emu.push = push

local pop = function()
	local val = read8(emu.sp) | (read8(emu.sp + 1) << 8)
	emu.sp = $ + 2
	return val
end
emu.pop = pop

local call = function(addr)
	emu.push(pc)
	pc = addr
end
emu.call = call

emu.callInt = function(int)
	inf = $ | int
end

emu.setPC = function(v)
	pc = v&0xFFFF
end

emu.addPC = function(v)
	pc = ($+v)&0xFFFF
end

emu.lockMem = function() memlock = true end
emu.unlockMem = function() memlock = false end
emu.ei = function() ime = true end
emu.di = function() ime = false end
emu.halt = function() stop = STOPHALT end
emu.stop = function() stop = STOPSTOP end

--Misc MMIO
local mmioDiv = function()
	return (divTic >> 6) -- Internal divider is offset so need to multiply this value
end
emu.mmioR[MMIODIV] = mmioDiv

local mmioDivW = function(_)
	divTic = 0
end
emu.mmioW[MMIODIV] = mmioDivW
emu.resetDIV = mmioDivW

local mmioInf = function()
	return inf
end
emu.mmioR[MMIOIF] = mmioInf

local mmioInfW = function(val)
	inf = 0xE0 | (val&0x1F)
end
emu.mmioW[MMIOIF] = mmioInfW

local mmioTimaR = function()
	return timer
end

local mmioTmaR = function()
	return timerMod
end

local mmioTacR = function()
	return timerCtrl
end

local mmioTmaW = function(val)
	timerMod = val
end

local mmioTacW = function(val)
	timerCtrl = 0xF8 | (val&0x7)
	currentFreq = timerClockFreq[(timerCtrl&TIMERCLOCK) + 1]
end

emu.mmioR[MMIOTIMA] = mmioTimaR
emu.mmioR[MMIOTMA] = mmioTmaR
emu.mmioR[MMIOTAC] = mmioTacR
emu.mmioW[MMIOTMA] = mmioTmaW
emu.mmioW[MMIOTAC] = mmioTacW


local step = function()
	local int
	local op = emu.op
	local e = emu
	local c
	while state == STATERUN and cpuCycle > 0 do
		--Test interrupts
		int = (inf & ine)&0x1F
		if int and stop then --If interrupt happens wake up if halted even if ime isn't set
			stop = 0
		end
		if int and ime then
			if int & 1 then --VBLANK
				inf = $ & 0xFE
				ime = false
				call(0x40)
			elseif int & 2 then --STAT
				inf = $ & 0xFD
				ime = false
				call(0x48)
			elseif int & 4 then --TIMER
				inf = $ & 0xFB
				ime = false
				call(0x50)
			elseif int & 8 then --SERIAL
				inf = $ & 0xF7
				ime = false
				call(0x58)
			elseif int & 16 then --JOYPAD
				inf = $ & 0xEF
				ime = false
				call(0x60)
			end
		end

		if not stop then --Don't run if stopped
			--Execute next instruction
			c = read8(pc)
			pc = $ + 1
			op[c](c)
		elseif stop == STOPHALT then
			nextCycle()
		elseif stop == STOPSTOP then --Do nothing but check if an input is pressed if stop(yes softlock if input select isn't set)
			if (~e.getInput())&0x0F then
				stop = 0
			else -- Don't hang
				return
			end
		end
	end
end

-- Commands/Hooks

--Debug function consisting of printing some registers. Should be called within a HUD hook
emu.drawCoreDump = function(v)
	local x = 304
	local y = 16
	v.drawString(x, y, string.format("AF: %04x",reg[emu.REGAF+1]) ,V_HUDTRANS, "right") y = $ + 8
	v.drawString(x, y, string.format("BC: %04x",reg[emu.REGBC+1]) ,V_HUDTRANS, "right") y = $ + 8
	v.drawString(x, y, string.format("DE: %04x",reg[emu.REGDE+1]) ,V_HUDTRANS, "right") y = $ + 8
	v.drawString(x, y, string.format("HL: %04x",reg[emu.REGHL+1]) ,V_HUDTRANS, "right") y = $ + 12
	v.drawString(x, y, string.format("SP: %04x",emu.sp) ,V_HUDTRANS, "right") y = $ + 8
	v.drawString(x, y, string.format("PC: %04x",pc) ,V_HUDTRANS, "right") y = $ + 16
	
	v.drawString(x, y, "IME: "..tostring(ime), V_HUDTRANS, "right") y = y + 8
	v.drawString(x, y, "INE: "..emu.hexString(ine,2), V_HUDTRANS, "right") y = y + 8
	v.drawString(x, y, "IMF: "..emu.hexString(inf,2), V_HUDTRANS, "right") y = y + 12
	v.drawString(x, y, "Bank: "..emu.hexString(emu.getBank(),3), V_HUDTRANS, "right") y = y + 16
	
	if not emu.getSRAMLock() then
		v.drawString(x, 184, "R", V_HUDTRANS)
	end
	
	v.drawString(x, y, "LY: "..emu.getLine(), V_HUDTRANS, "right") y = y + 8
	v.drawString(x, y, "STAT: "..emu.hexString(emu.getStat(),2), V_HUDTRANS, "right") y = y + 8
	v.drawString(x, y, "LCDC: "..emu.hexString(emu.getLCDC(),2), V_HUDTRANS, "right") y = y + 16
	
	v.drawString(x, y, "TAC: "..emu.hexString(timerCtrl,2), V_HUDTRANS, "right") y = y + 24
	
	v.drawString(x, y, string.format("Time: %d us. Speed: %d %%", dtime, (speedMul*100)/FRACUNIT), V_HUDTRANS, "right")
	
	x = 240
	y = 16
	v.drawString(x, y, "Stack:", V_HUDTRANS, "right")
	local s = emu.sp
	local val
	for i = 0, 15 do
		y = $ + 8
		if (s > 0xFF80) or (s >= 0xC000 and s < 0xFE00) then
			val = read16Raw(s)
			v.drawString(x, y, i..": "..emu.hexString(val, 4), V_HUDTRANS, "right")
		end
		s = $ + 2
	end
end

emu.drawStats = function(v)
	local x = 304
	local y = 16
	v.drawString(x, y, string.format("%d %%", (speedMul*100)/FRACUNIT), V_HUDTRANS, "right")
end

addHook("ThinkFrame", function()
	if state ~= STATERUN then return false end
	if cpuCycle > 0 then cpuCycle = 0 end
	speedMul = FRACUNIT -- Default to 1x
	if emu.speed.value then -- Yes define it there even if autoslow overwrite it. Used by the later actually to define speed cap
		speedMul = emu.speed.value
	end
	if autoslow.value then
		timeHistE = $ + 1
		--Change speed according to performance. Overly complicated yeah because of average
		if haveTime then
			if timeHistE>TICRATE then
				local s = timeHist/timeHistE
				timeHistE = 1
				timeHist = 0
				
				speedFac = FixedMul($, FixedDiv((autoslow.value >> 4) * FRACUNIT, (s >> 4) * FRACUNIT)) -- Reduce accuracy of this shit because it overflow otherwise
				if speedFac > speedMul then speedFac = speedMul end
			end
		else
			speedFac = FRACUNIT
		end
		speedMul = speedFac
	end
	cpuCycle = $ + ((cyclePerTic*(speedMul >> 4)) / 0x1000) -- Apply the speed factor. Just in case we accelerate the emulation or make GBC, reduce accuracy
	collectgarbage("stop") -- Disable the gc to slightly increase speed
	if cpuCycle >= 0 then
		local time = getTimeMicros()
		step()
		dtime = getTimeMicros() - time
		timeHist = $ + dtime
	end
	collectgarbage("collect")
end)

emu.coreSaveConfig = function(t)
	t["limiter"] = autoslow.value
	t["speed"] = emu.speed.value
end

emu.coreLoadConfig = function(t)
	if t["limiter"] then
		COM_BufInsertText(consoleplayer, string.format("emulimiter %s", t["limiter"]))
	end
	if t["speed"] and tonumber(t["speed"]) then
		COM_BufInsertText(consoleplayer, string.format("emuspeed %f", tonumber(t["speed"])))
	end
end

addHook("GameQuit", function()
	stopEmu(consoleplayer)
	emu.saveConfig()
end)

--Don't work I don't know why
--[[COM_AddCommand("emupause", function(_)
	if state == STATERUN then
		state = STATEERROR
	end
end, COM_LOCAL)

COM_AddCommand("emuresume", function(_)
	if state == emu.STATEERROR then
		state = STATERUN
	end
end, COM_LOCAL)]]

COM_AddCommand("emustep", function(p, n)
	if state < STATERUN then return end
	if n == nil then n = 1
	else n = tonumber($) end
	state = STATERUN
	cpuCycle = $ + n
	step()
	state = emu.STATEERROR --Pause emulation(but making it in error state yes)
end, COM_LOCAL)

local stopEmu = function(p)
	if p == consoleplayer and state ~= emu.STATEOFF then
		state = emu.STATEOFF
		emu.saveSRAM()
	end
	p.playemu = false
end

COM_AddCommand("emustart", function(p, r)
	stopEmu(p)
	if r == nil then
		CONS_Printf(p, "Usage: emustart <file>. \nFile being relative to the luafiles/client/emu directory.")
		return
	end
	if p == consoleplayer then
		if not emu.start(r) then --Cancel the start if it failed to start
			COM_BufInsertText(p, "emustop")
		end
	end
	p.playemu = true
end)

COM_AddCommand("emustop", stopEmu)